// --- [ TWI_Master.c ] -------------------------------------------------------
//
//       tab = 3 - 08.03.2015
//
//       Repeated Start eingebaut
//
//       Beachten: Ausreichender TWI-Buffer fr die bertragung aller Bytes.
//                 Die Anzahl der zu lesenden / zu schreibenden Bytes immer
//                 um 1 erhhen (fr die TWI-Adresse)
//                 Die Daten stehen dann im Buffer ab Adresse 1 !
//                 an Adresse 0 steht die TWI-Adresse.

/*****************************************************************************
*
* File              : TWI_Master.c - Weitgehend vernderte Datei !!
* Compiler          : WINAVR 3.4.6
*
* AppNote           : AVR315 - TWI Master Implementation
*
* Description       : This is a sample driver for the TWI hardware modules.
*                     It is interrupt driveren. All functionality is controlled through
*                     passing information to and from functions. Se main.c for samples
*                     of how to use the driver.
*
* nderung: in ISR(TWI) bei switch / default: -> Stop senden !!
* Zur Unterscheidung mit Slave-Bibliothek sind einige Funktionsnamen gendert
* die Union TWI_statusReg ist auf eine 'normale' Variable reduziert
* Der Funktionsaufruf 'TWI_Transceiver_Busy( void )' ist durch den Inhalt der
* Funktion 'while ( TWCR & (1<<TWIE) )' ersetzt.
* Separate Funktionen zum Schreiben und Lesen aufgebaut.
*
****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>

#include "TWI_Master.h"

#define TRUE            1
#define FALSE           0
#define REPEATED_START	TRUE

// --- lokale Daten --------------------------------------------------------
static uint8_t   TWI_MA_buf[TWI_BUF_SIZE];		// Transceiver buffer
static uint8_t   TWI_msgSize;                	// Number of bytes to be transmitted.
volatile uint8_t TWI_success = TRUE;

#if REPEATED_START == TRUE
volatile uint8_t repeatedStart = FALSE;          // Es soll anstelle eines STOP ein REPEATED START gesendet werden
#elif REPEATED_START == FALSE
#else
#warning --> Konstante REPEATED_START undefinied
#endif

// ----------------------------------------------------------------------------
/* ist als Makro definiert
uint8_t TWI_Get_State( void )
{
//	return ( TWI_state );                        // Return error state.
	return ( TWSR );
} */
// ----------------------------------------------------------------------------
// liefert den Erfolg einer (Schreib-) Aktion zurck
// (ein TWI_MA_Read liefert den Erfolg automatisch zurck)
uint8_t TWI_MA_Write_Success(void)
{
	while (TWI_BUSY);                          	// Wait until TWI is ready for next transmission.
	return(TWI_success);
}

#if REPEATED_START == TRUE
void TWI_MA_RepeatedStart(void)
{
	repeatedStart = TRUE;
}
#endif

// ----------------------------------------------------------------------------
void TWI_MA_Init(void)
{
   TWBR = TWI_TWBR;                             // Set bit rate register (Baudrate). Defined in header file.
// TWSR = TWI_TWPS;                             // Not used. Driver presumes prescaler to be 00.
   TWDR = 0xFF;                                 // Default content = SDA released.
   TWCR = (1<<TWEN)|                            // Enable TWI-interface and release TWI pins.
          (0<<TWIE)|(0<<TWINT)|                 // Disable Interupt.
          (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|      // No Signal requests.
          (0<<TWWC);
}

// ----------------------------------------------------------------------------------------
// Schreibt die Daten eines Arrays ber die TWI-Schnittstelle raus
// Parameter 1: Ein Pointer auf ein Array
//              Das erste Byte muss die Adresse des TWI-Slaves sein, linksbndig ausgerichtet,
//              das Bit.0 (Read/Write-Bit) ist ohne Bedeutung, es wird automatisch gesetzt
// Parameter 2: Die Anzahl der zu schreibenden Bytes zuzglich '1' fr die Adresse in Byte 0

void TWI_MA_Write( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

	while ( TWI_BUSY );                          // Wait until TWI is ready for next transmission.

   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0]  = (msg[0] & 0xFE);          	// Store slave address with Write-Bit setting.
                                                // Die Daten umkopieren nach TWI_MA_buf[]
   for ( i = 1; i < msgSize; i++ ) TWI_MA_buf[i] = msg[i];

	TWI_success = FALSE;
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);

	/* Falls der (Miss-) Erfolg der Aktion abgewartet werden und zurckgeliefert werden soll
	while ( TWI_BUSY );                          // Wait until TWI is ready for next transmission.
   return(TWI_success);
   */
}

// ----------------------------------------------------------------------------------------
// Schreibt die Daten eines Arrays ber die TWI-Schnittstelle raus
// Parameter 1: Ein Pointer auf ein Array
//              Das erste Byte muss die Adresse des TWI-Slaves sein, linksbndig ausgerichtet,
//              das Bit.0 (Read/Write-Bit) ist ohne Bedeutung, es wird automatisch gesetzt
// Parameter 2: Die Anzahl der zu lesenden Bytes zuzglich '1' fr die Adresse in Byte 0
//
// Das Byte.0 ist die TWI-Adresse, die Daten beginnen bei Byte.1 !!!
//

uint8_t TWI_MA_Read( uint8_t *msg, uint8_t msgSize )
{
   uint8_t i;

	while ( TWI_BUSY);                           // Wait until TWI is ready for next transmission.

   TWI_msgSize = msgSize;                      	// Number of data to transmit.
   TWI_MA_buf[0] = (msg[0] | 0x01);             // Store slave address with Read-Bit setting.

	TWI_success = FALSE;
   TWCR = (1<<TWEN)|                            // TWI Interface enabled.
          (1<<TWIE)|(1<<TWINT)|                 // Enable TWI Interupt and clear the flag.
          (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|      // Initiate a START condition.
          (0<<TWWC);

// Jetzt werden die Daten im Hintergrund abgeholt,
// wenn die bertragung abgeschlossen ist, das wird zuerst
// TWIE gelscht und gleichzeitig  TWSTO gesetzt - dadurch wird ein TWI-Stop generiert
// ist das TWI-Stop ausgefhrt, wird auf TWSTO gelscht

	while ( TWCR & (1<<TWIE | 1<<TWSTO) );       // Wait until TWI is ready for next transmission.

	if (TWI_success)            						// Last transmission competed successfully.
   {                                            // Die Daten wurden in der ISR in TWI_MA_buf[] eingelesen
      for ( i = 0; i < msgSize; i++ )           // Copy data from Transceiver buffer.
      {
         msg[i] = TWI_MA_buf[i];
      }
   }
   return( TWI_success );
}

// --------------------------------------------------------------------------------------
ISR(TWI_vect)
{
   static uint8_t idx;

   switch (TWSR)
   {
   case TWI_START:              						// START has been transmitted
   case TWI_REP_START:          						// Repeated START has been transmitted
      idx = 0;           						      // Set buffer pointer to the TWI Address location

// --- TWI Master Transmitter status codes ------------------------------------

   case TWI_MTX_ADR_ACK:        						// SLA+W has been transmitted and ACK received
   case TWI_MTX_DATA_ACK:       						// Data byte has been transmitted and ACK received
      if (idx < TWI_msgSize)
			{
			TWDR = TWI_MA_buf[idx++];			     // Adresse bzw. Daten nach TWDR schreiben
			TWCR = (1<<TWEN)|                     // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|           // Enable TWI Interupt and clear the flag to send byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)|
               (0<<TWWC);
			}

      else                                      // Send STOP after last byte
			{
			TWI_success = TRUE;       					// Set status bits to completed successfully.

#if REPEATED_START == TRUE
			if (repeatedStart)
				{


				TWCR = (1<<TWEN)|                   // TWI Interface enabled
					(0<<TWIE)|(0<<TWINT)|            // Disable TWI Interrupt and clear the flag
					(0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // do nothing
					(0<<TWWC);
				repeatedStart = FALSE;
				}
			else
				{
				TWCR = (1<<TWEN)|                   // TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
               (0<<TWWC);
				}
#elif REPEATED_START == FALSE

			TWCR = (1<<TWEN)|                   	// TWI Interface enabled
               (0<<TWIE)|(1<<TWINT)|            // Disable TWI Interrupt and clear the flag
               (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)| // Initiate a STOP condition.
               (0<<TWWC);


#endif // REPEATED_START
			}
      break;

// --- TWI Master Receiver status codes ---------------------------------------

   case TWI_MRX_DATA_ACK:                       // Data byte has been received and ACK transmitted
      TWI_MA_buf[idx++] = TWDR;				      // gelesene Daten in den Puffer schreiben

   case TWI_MRX_ADR_ACK:                        // SLA+R has been tramsmitted and ACK received
      if (idx < (TWI_msgSize -1) )               // Detect the last byte to NACK it.
			{
        TWCR = (1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (1<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send ACK after reception
               (0<<TWWC);
			}
      else                                      // Send NACK after next reception
			{
        TWCR = (1<<TWEN)|                       // TWI Interface enabled
               (1<<TWIE)|(1<<TWINT)|            // Enable TWI Interupt and clear the flag to read next byte
               (0<<TWEA)|(0<<TWSTA)|(0<<TWSTO)| // Send NACK after reception
               (0<<TWWC);
			}
      break;

   case TWI_MRX_DATA_NACK:                      // Data byte has been received and NACK transmitted
      TWI_MA_buf[idx] = TWDR;
		TWI_success = TRUE;         					// Set status bits to completed successfully.
      TWCR = (1<<TWEN)|                         // TWI Interface enabled
             (0<<TWIE)|(1<<TWINT)|              // Disable TWI Interrupt and clear the flag
             (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   // Initiate a STOP condition.
             (0<<TWWC);
		break;

//		alle anderen Flle laufen ber default:
//    case TWI_ARB_LOST:                        // Arbitration lost
//    case TWI_MTX_ADR_NACK:                    // SLA+W has been transmitted and NACK received
//    case TWI_MRX_ADR_NACK:                    // SLA+R has been transmitted and NACK received
//    case TWI_MTX_DATA_NACK:                   // Data byte has been transmitted and NACK received
//    case TWI_NO_STATE;                        // No relevant state information available; TWINT = 0
//    case TWI_BUS_ERROR:                       // Bus error due to an illegal START or STOP condition
   default:
                                                // Reset TWI Interface
      TWCR = (1<<TWEN)|                         // Enable TWI-interface and release TWI pins
             (0<<TWIE)|(1<<TWINT)|              // Disable Interupt   - nderung: clear flag
             (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|   // No Signal requests - nderung: Stop Condition senden
             (0<<TWWC);
   }
}

// --- [ end of file ] --------------------------------------------------------
