// --- [ main.c ] -------------------------------------------------------------
//
// tab = 3
//
// Hardeware: fr ATMega48 mit 3.6864 MHz
// Der Interrupt Output A der RTC ist mit INT0 am ATMega verbunden (Pullup nach VCC)
// Die Ausgabe erfolgt via Serielle Schnittstelle (bei 9600 Baud, 8, 1)
//
// ----------------------------------------------------------------------------

#include <avr/io.h>
#include <avr/interrupt.h>
#include <inttypes.h>
#include <stdlib.h>
#include <avr/pgmspace.h>
#include <util/delay.h>

#include "global.h"
#include "rtc.h"
#include "rs232.h"
#include "TWI_Master.h"

void print_time(void);
void irqHandler(void);

uint8_t time[8];				// hier steht die Zeit, die aus der RTC abgeholt wurde
uint8_t dow_mask = ALL;		// an welchen Wochentagen ist der Alarm aktiv ?
volatile uint8_t flag = FALSE;

// Das Alarmintervall in Minuten (>= 1 und <=30)
// hier: Alle 2 Minuten wird ein TimerAlarm ausgelst
#define WAIT 	2	// (60 % DELAY) muss == 0 sein
#define LIMIT	(60 - WAIT)
// Weiter unten wird der Periodic Interrupt auf 1 Minuten Intervall eingestellt.

int main(void)
{
	uint8_t data[10];

	rs232_init();
	puts_p(PSTR("start"));
	put_c(13);

	TWI_MA_Init();
	sei();							// Die TWI-Schnittstelle arbeitet IRQ-gesteuert

	rtc_printRegisters();		// Register vor der Initialisierung ausgeben
	rtc_init();
	rtc_printRegisters();		// Register nach der Initialisierung ausgeben

										// alle Alarmeinstellungen erstmal zurcksetzen
	rtc_setPeriodicIRQ(P_IRQ_OFF);// Periodic Alarm lschen
	rtc_clrAlarm_A();
	rtc_clrAlarm_B();

	data[0] = 53;					// Zeit einstellen, beginnend mit Sekunde
	data[1] = 55;					// Minute
	data[2] = 20;					// Stunde
	data[3] = MITTWOCH;			// aktueller Wochentag
	data[4] = 24;					// Tag
	data[5] = 12;					// Monat
	data[6] = 15;					// Jahr
	rtc_setTime(data);

	data[0] = 56;					// Alarm A setzen (Minute, Stunde, Dow_Mask)
	data[1] = 20;
	data[2] = dow_mask;
	rtc_setAlarm_A(data);

	data[0] = 30;					// Alarm B setzen (Minute, Stunde, Dow_Mask)
	data[1] = 20;
	data[2] = dow_mask;
	rtc_setAlarm_B(data);

	rtc_setPeriodicIRQ(P_IRQ_MINUTE); // Periodic Interrupt im Minutenabstand

	rtc_printRegisters();		// und noch einmal alle Register ausgeben

	EIMSK = (1<<INT0);			// INT0 als low-level Interrupt

	while (TRUE)
	{
//		put_c('+');	put_c(13);				// schauen, ob das Sleep funktioniert


		// erst ins Sleep gehen, wenn alle laufenden Aktionen abgeschlossen sind
		while(TWI_BUSY);
		_delay_ms(20);

		// ------------------------ sleep
		SMCR = (1<<SM1 | 1<<SE);

		asm volatile("sleep");
		SMCR = 0;

		if (flag)					// wird in der INT0-IRQ gesetzt
		{
			flag = FALSE;
			irqHandler();
			EIMSK = (1<<INT0);	// Interrupt wieder aktivieren
		}
	}
}

// ----------------------------------------------------------------------------
// Ein low-level an INT0 lst einen Interrupt aus
// der INT0 muss dann deaktiviert werden,
// bis das Flag gelscht ist und der Pin wieder auf high gegangen ist.
ISR(INT0_vect)
{
	flag = TRUE;
	EIMSK = 0;
}

// ----------------------------------------------------------------------------
// wird aufgerufen, wenn ein Interrupt der RTC empfangen worden ist
// gibt die Alarmflags und Hinweise zu gesetzten Alarmbits aus,
// schreibt die nchste Alarmzeit abwechselnd nach Alarm_A / Alarm_B
void irqHandler(void)
{
	uint8_t i;
	uint8_t data[10];

	i = rtc_getAlarmFlags();	// holt die Flags ab / lscht die Flags auf der RTC

	puts_p(PSTR("AlarmFlags :  "));
	print_hexbyte(i);
	put_c(':');
	put_c(' ');

	if (i & 0x01) puts_p(PSTR("Alarm_B "));
	if (i & 0x02) puts_p(PSTR("Alarm_A "));
	if (i & 0x04) puts_p(PSTR("Periodic_Interrupt "));

	print_time();

	// zuerst den nchsten Alarm einstellen
	if (time[MINUTE] < LIMIT)
	{
		data[0] = time[MINUTE] + WAIT;
		data[1] = time[STUNDE];
		data[2] = dow_mask;
	}

	else
	{
		data[0] = 0;
		if (time[STUNDE] < 24)
		{
			data[1] = time[STUNDE] + 1;
		}
		else data[STUNDE] = 0;
		data[STUNDE] = dow_mask;

	}
	// die Alarmzeit zurckschreiben
	if (i & 1)
	{
		rtc_setAlarm_A(data);
	}
	else if (i & 2)
	{
		rtc_setAlarm_B(data);
	}
	rtc_printRegisters();
}

// ----------------------------------------------------------------------------
void print_time(void)
{
	// Holt die Zeit via time[] aus der Uhr ab
	rtc_getTime(time);

	// gibt aus im Format Tag, Monat, Jahr, Stunde, Minute, Sekunde
	print_dezbyte(time[4]);
	put_c('.');
	print_dezbyte(time[5]);
	put_c('.');
	print_dezbyte(time[6]);
	put_c(' ');
	print_dezbyte(time[2]);
	put_c(':');
	print_dezbyte(time[1]);
	put_c(':');
	print_dezbyte(time[0]);
	put_c(13);
}

// --- [ eof ] ----------------------------------------------------------------
